"""

"""

import base64
import requests
# Debug & Logging
import logging


class MDSAuth:
    __slots__ = (
        "config",
        "oauth_token",
        "session",
        "headers",
        "authenticate",
        "custom_function",
    )

    def __init__(self, config, custom_function=None):
        """
        Initializes the class and the internal configuration
        :param dict config:
        :param function custom_function: A python function to run as a custom authentication
        """
        self.config = config
        self.custom_function = custom_function
        auth_type = self.config.get("auth_type", None)
        self.headers = None

        if auth_type:
            logging.debug(f"MDSAuth::__init__() Authentication method: {auth_type}")
            self.authenticate = {
                "oauth": self.mds_oauth,
                "bearer": self.mds_auth_token,
                "basic": self.mds_http_basic,
                "custom": self.mds_custom_auth,
            }.get(auth_type.lower(), None)

            if not self.authenticate:
                raise Exception(
                    f"MDSAuth::__init__() Invalid authentication method provided, auth_type: '{auth_type}'"
                )
        else:
            raise Exception(
                f"MDSAuth::__init__() No authentication method provided, auth_type: '{auth_type}'"
            )

    def mds_oauth(self):
        logging.debug("MDSAuth::mds_oauth() Running OAuth authentication")
        auth_data = self.config.get("auth_data", {})
        token_url = self.config.get("token_url", None)

        logging.debug("MDSAuth::mds_oauth() Making OAuth HTTP Request...")
        if token_url:
            response = requests.post(token_url, data=auth_data)
        else:
            raise Exception(
                "MDSAuth::mds_oauth() No token_url defined in the settings."
            )

        auth_token_res_key = self.config.get("auth_token_res_key", None)
        if auth_token_res_key:
            token = response.json().get(auth_token_res_key, None)
        else:
            raise Exception(
                "MDSAuth::mds_oauth() 'auth_token_res_key' is "
                "not defined in the config, usually set to 'jwt' or 'access_token'."
            )

        if token:
            logging.debug("MDSAuth::mds_oauth() Received token: %s[...]" % (token[:6]))
            self.headers = {"Authorization": f"Bearer {token}"}

            return self.headers
        else:
            raise Exception("MDSAuth::mds_oauth() Token could not be resolved.")

    def mds_auth_token(self):
        logging.debug("MDSAuth::mds_auth_token() Running Token authentication")
        self.headers = {"Authorization": f'Bearer {self.config["token"]}'}
        return self.headers

    def mds_http_basic(self):
        logging.debug("MDSAuth::mds_oauth() Running HTTP Basic authentication")
        auth_data = self.config.get("auth_data", None)
        if auth_data:
            username = auth_data.get("username", None)
            password = auth_data.get("password", None)
            encoded_creds = base64.b64encode(
                f"{username}:{password}".encode("utf-8")
            ).decode("utf-8")
            self.headers = {"Authorization": f"Basic {encoded_creds}"}
            return self.headers
        else:
            raise Exception("No credentials provided")

        return None

    def mds_auth_custom(self):
        logging.debug("MDSAuth::mds_oauth() Running Custom authentication")
        pass

    def mds_custom_auth(self):
        return self.custom_function(self.config)

    def _gather_oauth_token(self):
        logging.debug("MDSAuth::_gather_oauth_token() Gathering OAuth token")
        pass



from .MDSClientBase import MDSClientBase

# Debug & Logging
import logging

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(levelname)s - %(message)s"
)


class MDSClient020(MDSClientBase):
    version = "0.2.0"

    # Specify the name for each parameter:
    # param_schema = {
    #   param_name_in_configuration: param_name_in_mds_v0_2_0
    # }
    param_schema = {
        "start_time": "start_time",
        "end_time": "end_time",
        "bbox": "bbox",
        "device_id": "device_id",
        "vehicle_id": "vehicle_id",
    }

    def __init__(self, config):
        MDSClientBase.__init__(self, config)

    @staticmethod
    def _has_trips(data):
        """
        Returns True if the data contains trips
        :param dict data: The response data as provided by self.__request
        :return bool:
        """
        return len(data.get("payload", {}).get("data", {}).get("trips", [])) > 0

    @staticmethod
    def _has_data(data):
        """
        Returns True if data has any valid content.
        :param dict data: The response data as provided by self.__request
        :return bool:
        """
        return len(data.get("payload", {}).get("data", {})) > 0

    @staticmethod
    def _has_next_link(data):
        """
        Returns true if data contains a valid link to the next page
        :param dict data: The response data as provided by self.__request
        :return bool:
        """
        return True if data.get("payload", {}).get("links", {}).get("next", None) else False

    @staticmethod
    def _get_next_link(data):
        """
        Returns the link for the next page
        :param dict data: The response data as provided by self.__request
        :return bool:
        """
        return data.get("payload", {}).get("links", {}).get("next", None)

    def _get_response_version(self, data):
        """
        Returns the MDS version number as provided in the response body
        :param dict data: The response data as provided by self.__request
        :return str: The version number as provided by the mds endpoint response
        """
        return data.get("payload", {}).get("version", self.version)

    def get_trips(
        self,
        start_time,
        end_time,
        **kwargs,
    ):
        """
        Returns a JSON dictionary with a list of all
        :param int start_time: The start time in unix format
        :param int end_time: The end time in unix format
        :param str vehicle_id: (Optional) The vehicle ID
        :param str bbox: (Optional) Specify a bounding box (e.g., bbox="-122.4183,37.7758,-122.4120,37.7858")
        :param bool paging: (Optional) An override to paging. Set to True to enable it.
        :return dict:
        """
        logging.debug(
            "MDSClient020::get_trips() Getting trips: %s %s "
            % (start_time, end_time)
        )

        params = {
            **{
                "start_time": int(start_time),
                "end_time": int(end_time)
            },
            **kwargs
        }

        # Set the required URI parameters
        for key, value in params.items():
            self.params[self.param_schema[key]] = value

        # Out trips accumulator
        trips_accumulator = []
        # Contains our MDS endpoint with /trips path
        current_endpoint = f"{self.mds_endpoint}/trips"
        # A flag whose value is True if there is more data to download
        has_next_link = False

        # Start an endless loop
        while True:

            # 1. Make the HTTP Request
            data = self._request(
                mds_endpoint=current_endpoint,
                headers=self.headers,
                params=None if has_next_link else self.params,
            )

            # 2. Check if we have a next link
            has_next_link = self._has_next_link(data)

            # 3. If the data has trips, process them.
            if self._has_trips(data):
                # Gather the trips from `data`
                trips = data["payload"]["data"]["trips"]
                # Append the trips to the current list
                trips_accumulator += trips
                # Wipe out the current trips variable and start over
                trips = None

            # 4. Quit loop if not paging
            if self.paging is False:
                logging.debug("MDSClient020::get_trips() Paging set to False, stopping request...")
                break

            # 5. The `next` link becomes our new endpoint
            current_endpoint = self._get_next_link(data)

            # 6. If the endpoint is None, then quit loop
            if current_endpoint:
                logging.debug(
                    "MDSClient020::get_trips() Next link: %s" % current_endpoint
                )
            else:
                break

        # Return trips in this envelope:
        return {
            "version": self._get_response_version(data),
            "data": {
                "trips": trips_accumulator
            }
        }



from datetime import datetime, timedelta
import pytz


class MDSTimeZone:
    def __init__(
        self,
        date_time_now,
        time_zone=None,
        offset=86400,
        **kwargs
    ):
        """
        Class Constructor
        :param datetime date_time_now: The date time that is to be considered the reference time.
        :param str time_zone: The timezone to be applied to the date time object.
        :param int offset: In seconds, how far back to look for.
        :param dict kwargs: Any other additional parameters
        """

        if time_zone is not None:
            self.time_zone = pytz.timezone(time_zone)

        self.time_start = self.get_time(
            offset=offset,
            date_time_now=date_time_now
        )

        self.time_end = self.get_time(
            date_time_now=date_time_now
        )

        # Any other values or overrides
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_time(self, offset=0, date_time_now=datetime.now()):
        """
        Builds a time object
        :param int offset:The offset to be applied in seconds
        :param datetime date_time_now: The date time object to use, or default to now.
        :return:
        """

        # If the offset is positive, the delta is negative (go back in time)
        # Else, the delta is positive (we go forward in time)
        offset_delta = (
            date_time_now - timedelta(seconds=offset)
            if offset > 0
            else date_time_now + timedelta(seconds=offset)
        )

        # If the offset is 0, then assume the value of date_time_now
        time = (
            date_time_now if offset == 0 else offset_delta
        )

        # Build a timezone-aware date time
        return self.time_zone.localize(time)

    def get_time_start(self, utc=False, unix=False):
        """
        Generates the start timestamp for a query.
        :param bool utc: If True, returns the timestamp in UTC timezone.
        :param bool unix: If True, returns the timestamp in UNIX Epoch format.
        :return:
        """
        # Try to get the time_end attribute from this class, or assume zero
        tz_aware_time_start = getattr(self, "time_start", 0)
        # Use current timezone or UTC if indicated
        tz_aware_output = tz_aware_time_start.astimezone(pytz.UTC) if utc else tz_aware_time_start
        # Use current format, or Unix Epoch if indicated
        return tz_aware_output.timestamp() if unix else tz_aware_output

    def get_time_end(self, utc=False, unix=False):
        """
        Generates the end timestamp for a query.
        :param bool utc: If True, returns the timestamp in UTC timezone.
        :param bool unix: If True, returns the timestamp in UNIX Epoch format.
        :return:
        """
        # Try to get the time_end attribute from this class, or assume zero
        tz_aware_time_end = getattr(self, "time_end", 0)
        # Use current timezone or UTC if indicated
        tz_aware_output = tz_aware_time_end.astimezone(pytz.UTC) if utc else tz_aware_time_end
        # Use current format, or Unix Epoch if indicated
        return tz_aware_output.timestamp() if unix else tz_aware_output



import time
import requests
from requests.exceptions import Timeout

# Debug & Logging
import logging

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(levelname)s - %(message)s"
)


class MDSClientBase:

    __slots__ = (
        "config",
        "params",
        "headers",
        "param_schema",
        "mds_endpoint",
        "paging",
        "delay",
        "timeout",
        "max_attempts",
    )

    def __init__(self, config):
        self.config = config
        self.headers = {}
        self.params = {}
        self.mds_endpoint = self.config.get("mds_api_url", None)
        self.paging = self.config.get("paging", True)
        self.delay = self.config.get("delay", 0)
        self.timeout = self.config.get("interval", None)
        self.max_attempts = self.config.get("max_attempts", 3)

    @staticmethod
    def _build_response(response):
        """
        Builds a data payload to be given back to the client
        :param object response: As provided by requests.get
        :return dict: A parsed response data
        """

        # In the future, we may want to refactor this
        # to handle 301 and 302 redirect responses.
        status_code = response.status_code if hasattr(response, "status_code") else -1
        success = status_code == 200
        message = response.content if hasattr(response, "content") else "No response message provided."

        logging.debug(
            f"MDSClientBase::_build_response() status_code: {status_code}"
        )

        return {
            "status_code": status_code,
            "response": "success" if success else "error",
            "message": "success" if success else f"Error: {message}",
            "payload": response.json() if success else {},
        }

    def _request(self, mds_endpoint, **kwargs):
        """
        Makes an HTTP request
        :param str mds_endpoint: The URL endpoint to make the request to
        :param dict params: (Optional) URI Parameters to add to the request
        :param dict headers: (Optional) A dictionary of HTTP headers to pass to the request
        :return dict:
        """
        logging.debug(f"\nMDSClientBase::__request() Making request...")

        # Load our endpoint, parameters and headers
        mds_params = kwargs.get("params", {})
        mds_headers = kwargs.get("headers", {})

        # Manage our current attempt to make an HTTP request
        current_attempts = 0

        # Log our current values
        logging.debug(f"MDSClientBase::__request() Details:")
        logging.debug(f"MDSClientBase::__request() mds_endpoint: {mds_endpoint}")
        logging.debug(f"MDSClientBase::__request() mds_params: {mds_params}")
        logging.debug(f"MDSClientBase::__request() mds_headers: {mds_headers}\n")

        # We are going to try N times as specified in self.max_attempts
        while True:
            # Increase current attempt
            current_attempts += 1

            # Wait N seconds as specified in `self.delay`
            time.sleep(self.delay)

            # Let's try to make an HTTP request
            try:
                logging.debug(
                    "MDSClientBase::__request() Attempting request: %s/%s -- Timeout %s, Paging: %s, Delay: %s"
                    % (
                        current_attempts, self.max_attempts, self.timeout, self.paging, self.delay
                    )
                )
                # Make actual request
                response = requests.get(
                    mds_endpoint,
                    params=mds_params,
                    headers=mds_headers,
                    timeout=self.timeout,
                )
                # Build a data json response
                data = self._build_response(response)

            # There was an exception, timeout or otherwise:
            except Exception as e:
                logging.debug(
                    "MDSClientBase::__request() Exception detected: %s" % (str(e))
                )
                data = {
                    "status_code": -1,
                    "response": "error",
                    "message": f"Error: {str(e)}",
                    "payload": {},
                }

            success = data.get("response", "error") == "success"
            logging.debug(
                "MDSClientBase::__request() Reported status: %s" % success
            )

            # Check if we have an error
            if success:
                break
            else:
                # First, log the response error
                logging.debug(
                    "MDSClientBase::__request() Unable to make request: %s"
                    % (data.get("message", "No error message provided"))
                )
                # Check if we still have attempts left
                if current_attempts < self.max_attempts:
                    continue  # Try again in next iteration
                else:
                    # We need to stop the execution, it seems we have a problem
                    raise Exception(
                        "Max attempts reached (%s): could not fetch MDS data at endpoint '%s'"
                        % (self.max_attempts, mds_endpoint)
                    )

        return data

    def set_header(self, key, value):
        """
        Adds an HTTP header to the list
        :param str key: The name of the header
        :param str value: The value of the HTTP header
        """
        logging.debug(
            f"MDSClientBase::set_header() Set header k: '{key}', v: '{value}'"
        )
        self.headers[key] = value

    def render_settings(self, headers={}):
        """
        Compiles the headers and the parameters
        :param dict headers: (Optional) Adds any additional headers to the list (e.g., authentication headers)
        """
        # 1. Consolidate current headers and new headers
        logging.debug("MDSClientBase::render_settings() Rendering headers")
        self.headers = {**self.headers, **headers}

        # 2. Initialize Param Schema & Overrides
        logging.debug("MDSClientBase::render_settings() Rendering parameters")
        params_override = self.config.get("mds_param_override", None)
        if isinstance(params_override, dict):
            for key, value in params_override.items():
                self.param_schema[key] = value

    def get_headers(self):
        """
        Returns the current list of headers
        :return dict:
        """
        return self.headers

    def set_paging(self, paging):
        """
        Allows to override the paging configuration
        :param bool paging: The new paging configuration. True to enable paging.
        """
        self.paging = paging

    def set_delay(self, delay):
        """
        Allows to override the paging configuration
        :param int delay: The new delay setting in seconds.
        """
        self.delay = delay

    def set_timeout(self, timeout):
        """
        Allows to override the paging configuration
        :param int timeout: The new timeout in seconds.
        """
        self.timeout = timeout

    def set_max_attempts(self, max_attempts):
        """
        Allows to override the max_attempts configuration
        :param int max_attempts: The new max_attempts setting.
        """
        self.max_attempts = max_attempts



"""

"""
from .clients import *
from .MDSAuth import MDSAuth

# Debug & Logging
import logging
logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(message)s')


class MDSClient:
    __slots__ = (
        "config",
        "authenticated",
        "mds_headers",
        "auth_headers",
        "version",
        "provider",
        "custom_client",
        "auth_client",
        "mds_client",
        "custom_authentication",
    )

    def __init__(self, config={}, custom_authentication=None, **kwargs):
        """
        Constructor for this class
        :param dict config: A dictionary of properties.
        :param function custom_authentication: A python function to use for authentication
        :param dic kwargs: Any additional parameters passed to subclasses

        Parameters:
            :param str version: The version of the mds library to be loaded.
            :param dict config: The configuration to be passed to the
            :param str provider: The provider name or UUID
        """
        # Merge config and kwargs into a single dictionary
        self.config = {**config, **kwargs}

        # Try to find in the config the MDS version we are working with
        self.provider = self.config.get("provider", None)
        # Tries to find version in the config, or assumes 0.2.0
        self.version = self.config.get("version", "0.2.0")
        # Try to find the default_class (an MDS class override) or assume None
        self.custom_client = self.config.get("custom_client", None)
        # Assume the headers to be empty
        self.mds_headers = None
        self.auth_headers = None
        # Assume authenticated is False
        self.authenticated = False

        # Try to find a custom authentication function, assume None
        self.custom_authentication = custom_authentication

        # Initialize authentication client
        self.auth_client = MDSAuth(
            config=self.config, custom_function=self.custom_authentication
        )

        # Initialize MDS Client
        self.mds_client = self.load_mds_client(
            version=self.version, custom=self.custom_client,
        )(config=self.config)

        self._load_custom_headers()
        self._authenticate()

    @staticmethod
    def load_mds_client(version, custom=None):
        """
        Returns the class reference to be initialized later.
        :param str version: The version of MDS to initialize
        :param object custom: MDS Class override option
        :return object: The MDS class to be used
        """
        # Check for class override
        if custom is not None:
            return custom
        # Proceed with normal version check & load class
        else:
            return {
                "0.2.0": MDSClient020,
                "0.3.0": MDSClient030,
                "0.4.0": MDSClient040,
            }.get(version, custom)

    def _load_custom_headers(self):
        logging.debug(f"MDSClient::get_trips() Loading custom headers...")
        custom_headers = self.config.get("headers", {})
        for key, value in custom_headers.items():
            self.mds_client.set_header(key=key, value=value)

    def get_trips(self, start_time, end_time):
        """
        Returns the trips for the current client
        :param start_time:
        :param end_time:
        :return:
        """
        logging.debug(f"MDSClient::get_trips() Getting trips for start_time: {start_time}, end_time: {end_time} ")
        return self.mds_client.get_trips(
            start_time=start_time, end_time=end_time
        )

    def show_config(self):
        """
        logging.debugs the current version & configuration of the client
        :return:
        """
        logging.debug(f"MDSClient::show_config() Current MDS version loaded: {self.mds_client.version}")
        logging.debug(self.mds_client.config)

    def _authenticate(self):
        """
        It authenticates the client using the provided configuration
        :return:
        """
        logging.debug("MDSClient::authenticate() Generating headers...")
        self.auth_headers = self.auth_client.authenticate()
        logging.debug("MDSClient::authenticate() Checking headers...")
        if self.auth_headers:
            logging.debug("MDSClient::authenticate() Authentication succeeded...")
            self.authenticated = True

            self.mds_client.set_header(
                "Accept", f"application/vnd.mds.provider+json;version={self.version[:3]}"
            )
            self.mds_client.render_settings(headers=self.auth_headers)

            logging.debug("MDSClient::authenticate() Final headers: ")
            logging.debug(self.mds_client.get_headers())

        else:
            logging.debug("MDSClient::authenticate() Authentication failed")
